;================================================================
;
;  3 phase AC current
;  version 1.0
;  
;  Microchip MPLAB IDE compiler   /www.microchip.com/
;
;================================================================

#include <P16F690.inc>

 __CONFIG _CPD_OFF & _CP_OFF & _BOR_OFF & _MCLRE_OFF & _PWRTE_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT

rxlength  EQU .5
rxmaxbuf  EQU .8  
avgcmaxl  EQU .20 
DelR0tmr  EQU .18         ; ~302sec = 5min = 18, 117s = 7, 67s = 4, 16,78s = 1
showtmr   EQU .15 
TmrVal    EQU .250
lfchar    EQU .10
crchar    EQU .13
ctout     EQU .20 
cdiv      EQU .30 


 cblock 0x120             ; bank 2 (80 bytes)
  avgc3: 2*avgcmaxl
 endc

 cblock 0xA0              ; bank 1 (80 bytes)
  avgc1: 2*avgcmaxl
  avgc2: 2*avgcmaxl
 endc

 cblock 0x20              ; bank 0 (80 bytes)
  rxbuf: rxmaxbuf
  rxchar, COMTOUT
  CTMP0
  char1,char2,char3,char4,dots,ctmp,tchar,disptmr,disppos
  ADCHi01, ADCLo01, ADCHi02, ADCLo02, ADCHi03, ADCLo03
  ADCHi1, ADCLo1, ADCHi2, ADCLo2, ADCHi3, ADCLo3
  ADCHi01m, ADCLo01m, ADCHi02m, ADCLo02m, ADCHi03m, ADCLo03m
  tflags, T1tmr, T2tmr, T3tmr
  R00tmr, R01tmr, R10tmr, R11tmr, R20tmr, R21tmr, R30tmr, R31tmr
  flags,avgc0,avgcidx,avgcH,avgcL
  breg,creg
  zofscur, divnum, ADCSel
  XCurrent, MCurrent0, MCurrent0x
  MCurrent1, MCurrent1x, MCurrent2, MCurrent2x, MCurrent3, MCurrent3x
 endc

                   
TMPW       EQU 0x7C
TMPS       EQU 0x7D
TMPFSR     EQU 0x7E
TMPPCLATH  EQU 0x7F


; tflags
CurOver1   EQU 0x0
CurOver2   EQU 0x1
CurOver3   EQU 0x2
DoRelay1   EQU 0x3
DoRelay2   EQU 0x4
DoRelay3   EQU 0x5
xxxxxxxx   EQU 0x6
Tmeasure   EQU 0x7 

; flags
UseManual  EQU 0x0 
UseAvgc    EQU 0x1 
UseCorrTbl EQU 0x2
UseMinCurr EQU 0x3  
UseManADC  EQU 0x4 
UseManMiMa EQU 0x5 

; PORTA
CurVol1    EQU 0x0 
CurVol2    EQU 0x1 
CurVol3    EQU 0x2 
CurSetting EQU 0x3    
CurVolSet  EQU 0x4     
MinMaxS    EQU 0x5

; PORTB
ResIO1    EQU 0x4
RXLine    EQU 0x5
Relay     EQU 0x6
TXLine    EQU 0x7

; PORTC
Ch1       EQU 0x0 
Ch2       EQU 0x1          
Ch3       EQU 0x2 
Ch4       EQU 0x3 
Dot       EQU 0x4
ClkDt     EQU 0x5
DispRst   EQU 0x6        
ResIO2    EQU 0x7        


movfv macro src,dst
     movfw src
     movwf dst
 endm

movlv macro src,dst
     movlw src
     movwf dst
 endm

fwcall macro fw,callwhat
     movfw fw
     call callwhat
 endm

lwcall macro lw,callwhat
     movlw lw
     call callwhat
 endm

cmprxb1 macro const,jumpto
     movfw rxbuf+1
     sublw const
     btfss STATUS,Z		
	 goto  jumpto
 endm


;#define enable_one_phase
#define enable_current_dot
#define disable_cross_page_boundary  errorlevel -306
#define enable_cross_page_boundary  errorlevel +306
#define disable_bank_warn errorlevel -302
#define enable_bank_warn  errorlevel +302
#define use_prgmem_correct_table
;#define use_4MHz                         ; else use 8MHz


     ORG     0x2100       ; eeprom data

     ORG     0x0
     goto    Start

     ORG     0x4   
Interrupt          
     movwf TMPW
     swapf STATUS,W
     clrf  STATUS  
     movwf TMPS
     movfw FSR
     movwf TMPFSR
     movfw PCLATH
     movwf TMPPCLATH
     clrf  PCLATH  

     btfsc PIR1,RCIF
     goto  DoUsart 
     btfsc INTCON,RABIF
     goto  DOPCInt
     btfsc INTCON,INTF 
     goto  DoINT0
     btfsc INTCON,T0IF 
     goto  DoTimer0
     btfsc PIR1,TMR1IF 
     goto  DoTimer1

IEEnd                   
     movfw TMPPCLATH
     movwf PCLATH
     movfw TMPFSR
     movwf FSR
     swapf TMPS,W
     movwf STATUS
     swapf TMPW,F
     swapf TMPW,W
     retfie


DoUsart  
     movlv ctout,COMTOUT
     movlw rxlength
     subwf rxbuf,W
     btfsc STATUS,C       
     goto  DoUsartClrFIFO
     incf  rxbuf,F		  
     movlw rxbuf		  
     addwf rxbuf,W        
     movwf FSR            
     movfv RCREG,INDF   
     btfss PIR1,RCIF     
     goto  DoUsartEx
     incf  FSR,F
     movfv RCREG,INDF
     incf  rxbuf,F		 
DoUsartEx                 
     goto  IEEnd
DoUsartClrFIFO
     movfw RCREG
     movfw RCREG
     goto  DoUsartEx

DoTimer0
     bsf   tflags,Tmeasure
DoTimer00 
     call  UpdateDisplay4026
     movfw T1tmr
     btfss STATUS,Z
     decf  T1tmr,F
     movfw T2tmr
     btfss STATUS,Z
     decf  T2tmr,F
     movfw T3tmr
     btfss STATUS,Z
     decf  T3tmr,F
     bcf   INTCON,T0IF
     goto  IEEnd

DoTimer1 
     movfw disptmr
     btfss STATUS,Z
     decf  disptmr,F
     incf  R00tmr,F
     movfw R00tmr
     btfss STATUS,Z
     goto  dtmr1
     movfw R01tmr
     btfss STATUS,Z
     decf  R01tmr,F
dtmr1
     movfw COMTOUT     
     btfsc STATUS,Z
     goto  dtmr2
     decf  COMTOUT,F      
     goto  DoTimer1Ex
dtmr2
     movfw rxbuf       
     btfsc STATUS,Z		
	 goto  DoTimer1Ex      
     lwcall 'C',SendChar
     fwcall rxbuf,SendChar
     clrf  rxbuf
     bcf   RCSTA,CREN	
     bsf   RCSTA,CREN	
DoTimer1Ex
     bcf   PIR1,TMR1IF  
     goto  IEEnd

DOPCInt
     bcf  INTCON,RABIF  
     goto IEEnd

DoINT0                    
     bcf   INTCON,INTF  
     goto  IEEnd


SendChar                 
     bcf   STATUS,RP1    
     bcf   STATUS,RP0    
     btfss PIR1,TXIF     
     goto  $-1           
     movwf TXREG
     return

 disable_bank_warn

InitTimer0               
     bsf   STATUS,RP0    
     bcf   OPTION_REG,T0CS
     bcf   OPTION_REG,PSA
#ifdef use_4MHz
     bsf   OPTION_REG,PS0
     bcf   OPTION_REG,PS1
     bcf   OPTION_REG,PS2
#else
     bcf   OPTION_REG,PS0 
     bsf   OPTION_REG,PS1
     bcf   OPTION_REG,PS2 
#endif
     bcf   STATUS,RP0 
     bcf   INTCON,T0IF
     return

InitTimer1            
     bcf   STATUS,RP0 
#ifdef use_4MHz
     movlv b'00000100',T1CON 
#else
     movlv b'00010100',T1CON 
#endif
     bsf   STATUS,RP0  
     bsf   PIE1,TMR1IE 
     bcf   STATUS,RP0  
     bcf   PIR1,TMR1IF 
     return

InitUSART
     bcf   STATUS,RP0   
     movlv b'10010000',RCSTA
     bsf   STATUS,RP0   
     movlv b'00100000',TXSTA 
     clrf  SPBRGH
#ifdef use_4MHz
     movlv .25,SPBRG   
     bsf   TXSTA,BRGH
     bcf   BAUDCTL,BRG16 
#else
     movlv .51,SPBRG  
     bcf   TXSTA,BRGH
     bsf   BAUDCTL,BRG16
#endif
     bsf   PIE1,RCIE    
     bcf   STATUS,RP0   
     return
 

Start
     bsf   STATUS,RP0    
     movfw OSCCON
     iorlw b'01110000'   
     movwf OSCCON      
     bcf   STATUS,RP0    
     bsf   STATUS,RP1    
     movlv b'00000000',CM1CON0
     movwf CM2CON0
     movlv b'00001111',ANSEL 
     clrf  ANSELH
     bcf   STATUS,RP1    
     bsf   STATUS,RP0    
     movlv b'00100000',ADCON1 
     bcf   STATUS,RP0     
     movlv b'00000000',ADCON0
     bsf   STATUS,RP0     
     movlv b'00111111',TRISA  
     movlv b'00100000',TRISB  
     movlv b'00000000',TRISC  
     bcf   STATUS,RP0     
     clrf  PORTC
     clrf  PORTA
     clrf  PORTB
     clrf  T1tmr
     clrf  T2tmr
     clrf  T3tmr
     clrf  tflags
     clrf  flags
     call  InitTimer1
     call  InitTimer0
     call  InitUSART
     movlv b'11000000',INTCON 
     bsf   T1CON,TMR1ON  
     bsf   INTCON,T0IE   
     call  ClearVars
     clrf  COMTOUT
     clrf  rxbuf
     movlv showtmr,disptmr
     clrf  disppos    
     movlw .9
     movwf char1
     movwf char2
     movwf char3
     movwf char4
     clrf  dots
     movlv b'10000001',ADCON0
     bsf   STATUS,RP0 
     movlv b'00001101',WDTCON
     bcf   STATUS,RP0     
 enable_bank_warn
 
Loop

     clrwdt
     call  CheckUSART
     btfss tflags,Tmeasure 
     goto  LoopCont
     bcf   tflags,Tmeasure   
     btfsc flags,UseManMiMa   
     goto  DoChk
     bcf   flags,UseMinCurr
     btfsc PORTA,MinMaxS      
     bsf   flags,UseMinCurr
DoChk
     call  CheckCurrent     
     btfss flags,UseMinCurr 
     goto  DoChkMax
DoChkMin
     call  IsBelow1
#ifndef enable_one_phase
     call  IsBelow2
     call  IsBelow3
#endif
     goto  ExMinMax
DoChkMax
     call  IsOver1          
#ifndef enable_one_phase
     call  IsOver2
     call  IsOver3
#endif
ExMinMax
     btfsc flags,UseManual 
     goto  CurSet    
     movlw (1 << DoRelay1) + (1 << DoRelay2) + (1 << DoRelay3)
     andwf tflags,W
     btfsc STATUS,Z
     goto  DoChk0
     call  EnableRelay1
     goto  DoChk1
DoChk0
     call  DisableRelay1
DoChk1
CurSet
     btfss PORTA,CurSetting
     goto  ShowDisp0          
     clrf  disppos
     movlv showtmr,disptmr
     movlv .0,char1           
     movlv .4,dots             

     fwcall MCurrent0,ConvertByteToChars
     movfv char3,char2
     movfv char4,char3
     movfv MCurrent0x,char4
     goto  LoopCont
ShowDisp0
     movfw disptmr
     btfss STATUS,Z
     goto  ShowDisp3
     movlv showtmr,disptmr
     bcf   STATUS,C
     rlf   disppos,F   
     movlw b'00000111'
     andwf disppos,W
     btfsc STATUS,Z 
     movlv b'00000001',disppos
     movlv .0,char2
     clrf  dots
     btfss disppos,0
     goto  ShowDisp1
     fwcall MCurrent1,ConvertByteToChars
     movlv .1,char1
#ifdef enable_current_dot
     movlv .4,dots              
     movfv char3,char2
     movfv char4,char3
     movfv MCurrent1x,char4
#endif
     goto  ShowDisp3
ShowDisp1
     btfss disppos,1
     goto  ShowDisp2
     fwcall MCurrent2,ConvertByteToChars
     movlv .2,char1
#ifdef enable_current_dot
     movlv .4,dots             
     movfv char3,char2
     movfv char4,char3
     movfv MCurrent2x,char4
#endif
     goto  ShowDisp3
ShowDisp2
     btfss disppos,2
     goto  ShowDisp3
     fwcall MCurrent3,ConvertByteToChars
     movlv .3,char1
#ifdef enable_current_dot
     movlv .4,dots            
     movfv char3,char2
     movfv char4,char3
     movfv MCurrent3x,char4
#endif
ShowDisp3

LoopCont
      goto  Loop


IsBelow1
     btfsc tflags,CurOver1
     goto  IsBelow11
     movlv TmrVal,T1tmr
IsBelow11
     bcf   tflags,CurOver1
     movfw ADCHi01
     subwf ADCHi1,W
     btfsc STATUS,Z
     goto  IsBelow12
     btfss STATUS,C
     bsf   tflags,CurOver1
     goto  IsBelow13
IsBelow12
     movfw ADCLo01
     subwf ADCLo1,W
     btfss STATUS,C
     bsf   tflags,CurOver1
IsBelow13
     bcf   tflags,DoRelay1
     btfss tflags,CurOver1
     return
     movfw T1tmr
     btfsc STATUS,Z
     bsf   tflags,DoRelay1
     return

IsBelow2
     btfsc tflags,CurOver2
     goto  IsBelow21
     movlv TmrVal,T2tmr
IsBelow21
     bcf   tflags,CurOver2
     movfw ADCHi01
     subwf ADCHi2,W
     btfsc STATUS,Z
     goto  IsBelow22
     btfss STATUS,C
     bsf   tflags,CurOver2
     goto  IsBelow23
IsBelow22
     movfw ADCLo01
     subwf ADCLo2,W
     btfss STATUS,C
     bsf   tflags,CurOver2
IsBelow23
     bcf   tflags,DoRelay2
     btfss tflags,CurOver2
     return
     movfw T2tmr
     btfsc STATUS,Z
     bsf   tflags,DoRelay2
     return

IsBelow3
     btfsc tflags,CurOver3
     goto  IsBelow31
     movlv TmrVal,T3tmr
IsBelow31
     bcf   tflags,CurOver3
     movfw ADCHi01
     subwf ADCHi3,W
     btfsc STATUS,Z
     goto  IsBelow32
     btfss STATUS,C
     bsf   tflags,CurOver3
     goto  IsBelow33
IsBelow32
     movfw ADCLo01
     subwf ADCLo3,W
     btfss STATUS,C
     bsf   tflags,CurOver3
IsBelow33
     bcf   tflags,DoRelay3
     btfss tflags,CurOver3
     return
     movfw T3tmr
     btfsc STATUS,Z
     bsf   tflags,DoRelay3
     return

IsOver1
     btfsc tflags,CurOver1
     goto  IsOver11
     movlv TmrVal,T1tmr
IsOver11
     bcf   tflags,CurOver1
     movfw ADCHi1
     subwf ADCHi01,W
     btfsc STATUS,Z
     goto  IsOver12
     btfss STATUS,C
     bsf   tflags,CurOver1
     goto  IsOver13
IsOver12
     movfw ADCLo1
     subwf ADCLo01,W
     btfss STATUS,C
     bsf   tflags,CurOver1
IsOver13
     bcf   tflags,DoRelay1
     btfss tflags,CurOver1
     return
     movfw T1tmr
     btfsc STATUS,Z
     bsf   tflags,DoRelay1
     return

IsOver2
     btfsc tflags,CurOver2
     goto  IsOver21
     movlv TmrVal,T2tmr
IsOver21
     bcf   tflags,CurOver2
     movfw ADCHi2
     subwf ADCHi01,W
     btfsc STATUS,Z
     goto  IsOver22
     btfss STATUS,C
     bsf   tflags,CurOver2
     goto  IsOver23
IsOver22
     movfw ADCLo2
     subwf ADCLo01,W
     btfss STATUS,C
     bsf   tflags,CurOver2
IsOver23
     bcf   tflags,DoRelay2
     btfss tflags,CurOver2
     return
     movfw T2tmr
     btfsc STATUS,Z
     bsf   tflags,DoRelay2
     return

IsOver3
     btfsc tflags,CurOver3
     goto  IsOver31
     movlv TmrVal,T3tmr
IsOver31
     bcf   tflags,CurOver3
     movfw ADCHi3
     subwf ADCHi01,W
     btfsc STATUS,Z
     goto  IsOver32
     btfss STATUS,C
     bsf   tflags,CurOver3
     goto  IsOver33
IsOver32
     movfw ADCLo3
     subwf ADCLo01,W
     btfss STATUS,C
     bsf   tflags,CurOver3
IsOver33
     bcf   tflags,DoRelay3
     btfss tflags,CurOver3
     return
     movfw T3tmr
     btfsc STATUS,Z
     bsf   tflags,DoRelay3
     return

DisableRelay1              
     movfw R01tmr
     btfsc STATUS,Z       
     bcf   PORTB,Relay    
     return

EnableRelay1               
     clrf  R00tmr          
     movlv DelR0tmr,R01tmr
     bsf   PORTB,Relay     
     return

ClearVars
     clrf  ADCSel
     call  ClearAvgc
     clrf  avgcH
     clrf  avgcL
     clrf  MCurrent0
     clrf  MCurrent0x
     clrf  MCurrent1
     clrf  MCurrent2
     clrf  MCurrent3
     bsf   flags,UseCorrTbl     
     movlv cdiv,divnum      
     movlv .0,zofscur
     clrf  ADCHi01m
     clrf  ADCLo01m
     return


CheckRangeLo
     movlw .4         
     subwf avgcH,W
     btfss STATUS,C
     return
     clrf  avgcL
     clrf  avgcH
     return

CheckRangeHi
     movlw .4         
     subwf avgcH,W
     btfss STATUS,C
     return
     movlv .3,avgcH
     movlv .255,avgcL
     return


CheckCurrent     
     incf  avgcidx,F 
     movlw avgcmaxl
     subwf avgcidx,W
     btfsc STATUS,C
     clrf  avgcidx
     call  ReadADC            
     movlv b'10000101',ADCON0 
     movfv ADCHi01,avgcH     
     movfv ADCLo01,avgcL
     movlv avgc1,avgc0       
     movlv .1,ADCSel
     call  CalcCurrent       
     movfv XCurrent,MCurrent1
     movfv DivRem,MCurrent1x
     movfv avgcH,ADCHi1      
     movfv avgcL,ADCLo1
     call  ReadADC            
     movlv b'10001001',ADCON0 
     movfv ADCHi01,avgcH
     movfv ADCLo01,avgcL
     movlv avgc2,avgc0
     movlv .2,ADCSel
     call  CalcCurrent
     movfv XCurrent,MCurrent2
     movfv DivRem,MCurrent2x
     movfv avgcH,ADCHi2
     movfv avgcL,ADCLo2
     call  ReadADC               
     movlv b'10001101',ADCON0    
     movfv ADCHi01,avgcH
     movfv ADCLo01,avgcL
; errorlevel -202
     movlv avgc3,avgc0
; errorlevel +202
     bsf   STATUS,IRP             
     movlv .3,ADCSel
     call  CalcCurrent
     bcf   STATUS,IRP             
     movfv XCurrent,MCurrent3
     movfv DivRem,MCurrent3x
     movfv avgcH,ADCHi3
     movfv avgcL,ADCLo3
     call  ReadADC                
     movlv b'10000001',ADCON0     
     btfss flags,UseManADC
     goto  CheckCurrent0
     movfv ADCHi01m,ADCHi01
     movfv ADCLo01m,ADCLo01
CheckCurrent0
     movfv ADCHi01,avgcH
     movfv ADCLo01,avgcL
     call  SetCurrentFromADC0
     return

CalcCurrent
     movfv avgcidx,creg      
     bcf   STATUS,C
     rlf   creg,W            
     addwf avgc0,w
     movwf FSR   
     movfv avgcH,INDF
     incf  FSR,F
     movfv avgcL,INDF
     movlv avgcmaxl,creg
     movfv avgc0,FSR
     clrf  avgcH
     clrf  avgcL
     clrf  AddHi
     clrf  AddLo
     btfss flags,UseAvgc         
     goto  detectpeak
avgcurrent
     movfv INDF,AddHi0
     incf  FSR,F
     movfv INDF,AddLo0
     incf  FSR,F
     call  Add16
     decfsz creg,F
     goto  avgcurrent    
     movfv AddHi,DivHi
     movfv AddLo,DivLo
     lwcall avgcmaxl,Div16          
     movfv DivHi,avgcH
     movfv DivLo,avgcL
     goto  CalcCurrent0
detectpeak
     movfv INDF,DivHi
     incf  FSR,F
     movfv INDF,DivLo
     incf  FSR,F
     movfw avgcH    
     subwf DivHi,W
     btfss STATUS,C           
     goto  detectpeak1
     btfss STATUS,Z
     goto  detectpeak2        
     movfw avgcL
     subwf DivLo,W
     btfss STATUS,C           
     goto  detectpeak1
detectpeak2
     movfv DivHi,avgcH
     movfv DivLo,avgcL
detectpeak1                   
     decfsz creg,F
     goto  detectpeak         
CalcCurrent0
     call  SetCurrentFromADC  


CalcCorrect
     btfss flags,UseCorrTbl
     goto  contcorr
 disable_cross_page_boundary
#ifdef use_prgmem_correct_table
     movlw .1
     subwf ADCSel,W
     btfsc STATUS,Z
     goto  CalcCorrect1
     movlw .2
     subwf ADCSel,W
     btfsc STATUS,Z
     goto  CalcCorrect2
     movlw .3
     subwf ADCSel,W
     btfsc STATUS,Z
     goto  CalcCorrect3
     goto  CalcCorrect0
CalcCorrect1
     movlv HIGH corrtable1,PCLATH
     movfw XCurrent
     pagesel corrtablecompute1
     call corrtablecompute1
     pagesel CalcCorrect
     goto CalcCorrect0
CalcCorrect2
     movlv HIGH corrtable2,PCLATH
     movfw XCurrent
     pagesel corrtablecompute2
     call corrtablecompute2
     pagesel CalcCorrect
     goto CalcCorrect0
CalcCorrect3
     movlv HIGH corrtable3,PCLATH
     movfw XCurrent
     pagesel corrtablecompute3
     call corrtablecompute3
     pagesel CalcCorrect
     goto CalcCorrect0
#else
#endif
 enable_cross_page_boundary
CalcCorrect0
     movwf breg
     btfsc breg,7
     goto subcorr
     addwf avgcL,F
     btfsc STATUS,C
     incf  avgcH,F
     call  CheckRangeHi
     goto contcorr
subcorr
     andlw b'01111111'
     subwf avgcL,F
     btfss STATUS,C
     decf  avgcH,F
     call  CheckRangeLo
contcorr
     movfw zofscur            
     subwf avgcL,F
     btfss STATUS,C
     decf  avgcH,F
     call  CheckRangeLo
     return


RoundADC
     bcf   STATUS,C
     rrf   divnum,W
     subwf DivRem,w
     btfss STATUS,C
     return
     incf  DivLo,F
     btfss STATUS,Z
     return
     movlv .255,DivLo
     return

SetCurrentFromADC
#ifdef enable_current_dot
     movfv avgcL,DivLo
     movfv avgcH,DivHi 
     fwcall divnum,Div16
     movfw DivLo
     movwf XCurrent
     movfv DivRem,MulLo
     clrf  MulHi 
     lwcall .10,Mul16
     movfv MulHi,DivHi
     movfv MulLo,DivLo
     fwcall divnum,Div16
     movwf DivRem
#else
     movfv avgcL,DivLo
     movfv avgcH,DivHi 
     fwcall divnum,Div16
     call  RoundADC
     movfw DivLo
     movwf XCurrent      
#endif
     return                         

SetCurrentFromADC0
     movfv avgcL,DivLo
     movfv avgcH,DivHi 
     fwcall divnum,Div16
     movfw DivLo
     movwf MCurrent0
     movfv DivRem,MulLo
     clrf  MulHi 
     lwcall .10,Mul16
     movfv MulHi,DivHi
     movfv MulLo,DivLo
     fwcall divnum,Div16
     movwf MCurrent0x            
     return                      



ClearAvgc
     clrf  avgcH
     clrf  avgcL
     movlv avgcmaxl,creg        
     movwf avgcidx
     movlv avgc1,FSR
     call  clravg
     movlv avgcmaxl,creg        
     movlv avgc2,FSR
     call  clravg
     movlv avgcmaxl,creg        
     bsf   STATUS,IRP
; errorlevel -202
     movlv avgc3,FSR
; errorlevel +202
     call  clravg
     bcf   STATUS,IRP
     return
clravg
     clrf  INDF              
     incf  FSR,F
     clrf  INDF              
     incf  FSR,F
     decfsz creg,F
     goto  clravg
     return

ConvertByteToChars
     movwf DivLo
     movlv .0,breg
     lwcall .10,Div8
     btfss STATUS,Z
     movwf breg
     movfv DivRem,char4 
     movfv breg,char3
     return


UpdateDisplay4026    
     bcf   PORTC,Ch1
     bcf   PORTC,Ch2
     bcf   PORTC,Ch3
     bcf   PORTC,Ch4
     bcf   PORTC,Dot
     btfsc tchar,0
     goto  UpdateDisplay4026_1
     btfsc tchar,1
     goto  UpdateDisplay4026_2
     btfsc tchar,2
     goto  UpdateDisplay4026_3
     goto  UpdateDisplay4026_4
UpdateDisplay4026_1
     movfw char1
     btfsc STATUS,Z
     goto  UpdateDisplay4026_1x  
     call  Update4026
     btfsc dots,0
     bsf   PORTC,Dot
     bsf   PORTC,Ch1
UpdateDisplay4026_1x
     movlv .2,tchar
     return
UpdateDisplay4026_2
     movfw char2
     btfsc STATUS,Z
     goto  UpdateDisplay4026_2x  
     call  Update4026
     btfsc dots,1
     bsf   PORTC,Dot
     bsf   PORTC,Ch2
UpdateDisplay4026_2x
     movlv .4,tchar
     return
UpdateDisplay4026_3
     movfw char3
     call  Update4026
     btfsc dots,2
     bsf   PORTC,Dot
     bsf   PORTC,Ch3
     movlv .8,tchar
     return
UpdateDisplay4026_4
     movfw char4
     call  Update4026
     btfsc dots,3
     bsf   PORTC,Dot
     bsf   PORTC,Ch4
     movlv .1,tchar
     return

Update4026   
     bcf  PORTC,ClkDt  
     bsf  PORTC,DispRst
     nop
     nop
     bcf  PORTC,DispRst
     movwf ctmp
     btfsc STATUS,Z
     return
Update4026x
     bsf  PORTC,ClkDt
     nop
     nop
     bcf  PORTC,ClkDt   
     decfsz ctmp,f
     goto Update4026x
     return


CheckUSART
     movlw rxlength
     subwf rxbuf,W
     btfss STATUS,C      
     return
     bcf   RCSTA,CREN    
     movfv rxbuf,creg
     clrf  rxbuf         
cmd0    
     cmprxb1 .0,cmd1     
     call  SendBuffer    
     goto  CheckUSARTEx
cmd1
     cmprxb1 .1,cmd2
     call  SendVersion
     goto  CheckUSARTEx
cmd2
     cmprxb1 .2,cmd3
     call  SendCRLF      
     goto  $             
     goto  Start         
     goto  CheckUSARTEx
cmd3
     cmprxb1 .3,cmd4     
     movfw rxbuf+2
     btfss STATUS,Z
     bsf   PORTB,Relay   
     movfw rxbuf+2
     btfsc STATUS,Z
     bcf   PORTB,Relay   
     goto  CheckUSARTEx
cmd4
     cmprxb1 .4,cmd5
     goto  cmd5_4
;     goto  CheckUSARTEx
cmd5
     cmprxb1 .5,cmd6
     fwcall PORTC,SendChar
     fwcall PORTB,SendChar
     fwcall PORTA,SendChar
     fwcall flags,SendChar
     fwcall tflags,SendChar
     fwcall zofscur,SendChar
     fwcall MCurrent0,SendChar
     fwcall MCurrent0x,SendChar
     fwcall MCurrent1,SendChar
     fwcall MCurrent2,SendChar
     fwcall MCurrent3,SendChar
cmd5_4
     fwcall ADCLo01m,SendChar
     fwcall ADCHi01m,SendChar
     fwcall ADCLo01,SendChar
     fwcall ADCHi01,SendChar
     fwcall ADCLo1,SendChar  
     fwcall ADCHi1,SendChar
     fwcall ADCLo2,SendChar
     fwcall ADCHi2,SendChar
     fwcall ADCLo3,SendChar
     fwcall ADCHi3,SendChar
     goto  CheckUSARTEx
cmd6
     cmprxb1 .6,cmd7
     movfw rxbuf+2
     btfss STATUS,Z
     bsf   flags,UseManADC  
     movfw rxbuf+2
     btfsc STATUS,Z
     bcf   flags,UseManADC  
     movfw rxbuf+3          
     movwf ADCHi01m
     movfw rxbuf+4          
     movwf ADCLo01m
     goto  CheckUSARTEx
cmd7
     cmprxb1 .7,cmd8
     movfw rxbuf+2
     btfss STATUS,Z
     bsf   flags,UseManual  
     movfw rxbuf+2
     btfsc STATUS,Z
     bcf   flags,UseManual  
     movfw rxbuf+3
     btfss STATUS,Z
     bsf   flags,UseMinCurr 
     movfw rxbuf+3
     btfsc STATUS,Z
     bcf   flags,UseMinCurr 

     movfw rxbuf+4
     btfss STATUS,Z
     bsf   flags,UseManMiMa 
     movfw rxbuf+4
     btfsc STATUS,Z
     bcf   flags,UseManMiMa 
     goto  CheckUSARTEx
cmd8
     cmprxb1 .8,cmd9
     goto  CheckUSARTEx
cmd9
     cmprxb1 .9,cmd10
     goto  CheckUSARTEx
cmd10
     lwcall 'N',SendChar
     lwcall 'I',SendChar
CheckUSARTEx
     call  SendCRLF
     clrf  rxbuf
     bsf   RCSTA,CREN    
     return


 disable_bank_warn
ReadADC    
     bsf   ADCON0,GO_DONE 
IGetADCData0
     btfsc ADCON0,GO_DONE
     goto  IGetADCData0
     movfv ADRESH,ADCHi01
     bsf   STATUS,RP0     
     movfw ADRESL
     bcf   STATUS,RP0     
     movwf ADCLo01
     bcf   PIR1,ADIF
     return
 enable_bank_warn


SendBuffer
     fwcall creg,SendChar
     movlv rxbuf,FSR   
SendBuffer1
     incf  FSR,F
     fwcall INDF,SendChar
     decfsz creg,F
     goto  SendBuffer1
     return

SendVersion
     clrf  CTMP0
SendVersion1
     movlv HIGH version,PCLATH
     fwcall CTMP0,version
     andlw 0xFF
     btfsc STATUS,Z
     return
     call  SendChar        
     incf  CTMP0,F
     goto  SendVersion1

SendCRLF
     lwcall crchar,SendChar
     lwcall lfchar,SendChar
     return


 cblock
  MathLoopCnt,MthTmp0,MthTmp1
  DivLo,DivHi,DivRem,DivD
  MulLo,MulHi,MulHih,MulPl,MulPx
  AddHi,AddLo,AddHi0,AddLo0
 endc

Add16
     movfw AddLo0
     addwf AddLo,F
     movfw AddHi0
     btfsc STATUS,C
     goto  Add16_0     
     addwf AddHi,F
     return
Add16_0
     addwf AddHi,F
     incf  AddHi,F
     return

Mul16
     movwf MulPx
     movfw MulLo
     movwf MthTmp0
     movfw MulHi
     movwf MthTmp1
     clrf  MulHih
     clrf  MulHi
     clrf  MulLo
     bsf   MulLo,7
Mul16_0
     rrf   MulPx,F
     skpc
     goto Mul16_1
     movfw MthTmp0
     addwf MulHi,F
     movfw MthTmp1
     skpnc
     incfsz MthTmp1,W
     addwf MulHih,F
Mul16_1
     rrf MulHih,F
     rrf MulHi,F
     rrf MulLo,F
     skpc
     goto Mul16_0
     return


Div8
     movwf DivD
     clrf  DivRem
     movlw .8
     movwf MathLoopCnt
Div8_0
     rlf   DivLo,W
     rlf   DivRem,F
     movfw DivD
     subwf DivRem,F
     btfsc STATUS,C
     goto  Div8_1          
     addwf DivRem,F
     bcf   STATUS,C
Div8_1
     rlf   DivLo,F
     decfsz MathLoopCnt,F
     goto  Div8_0
     movfw DivLo
     return

Div16
     clrf  DivRem
     movwf DivD
     addlw .0
     btfsc STATUS,Z    
     goto  Div16_4
     movlw .8
     movwf MathLoopCnt
Div16_0      
     rlf   DivHi,W
     rlf   DivRem,F
     movfw DivD
     subwf DivRem,F
     btfsc STATUS,C
     goto  Div16_1          
     addwf DivRem,F
     bcf   STATUS,C
Div16_1
     rlf   DivHi, F
     decfsz MathLoopCnt,F
     goto  Div16_0
     clrf  MthTmp0
     movlw .8
     movwf MathLoopCnt
Div16_2
     rlf   DivLo,W
     rlf   DivRem, F
     rlf   MthTmp0, F
     movfw DivD
     subwf DivRem, F
     clrf  MthTmp1
     clrw
     btfss STATUS,C
     incfsz MthTmp1,W
     subwf MthTmp0,F
     btfsc STATUS,C
     goto  Div16_3
     movfw DivD
     addwf DivRem,F
     clrf  MthTmp1
     clrw
     btfsc STATUS,C
     incfsz MthTmp1,W
     addwf MthTmp0,F
     bcf   STATUS,C
Div16_3
     rlf   DivLo,F
     decfsz MathLoopCnt,F
     goto  Div16_2
     movfw DivLo
     bcf   STATUS,Z
     return
Div16_4                
     clrw
     clrf  DivHi
     clrf  DivLo
     bsf   STATUS,Z
     return



     ORG 0x7E0
version   
     addwf PCL,F
     dt "3 phase AC current v1.0",0



     ORG 0x8FF              ; table is on prg page 2
#include corrtable1.inc

     ORG 0xAFF              ; table is on prg page 2
#include corrtable2.inc

     ORG 0xCFF              ; table is on prg page 2
#include corrtable3.inc


 END

